/*
 *    Copyright  1996, 1997, 1998, 1999, 2000 by Remco Treffkorn
 *    Distributed under the terms of the Q Public License. Please see
 *    the included file LICENSE.
 */
/******************************************************************************
 *  label.c   labelling routines                                              *
 ******************************************************************************/
#include <config.h>
#include <X11/Xlib.h>
#include <stdio.h>
#include <netinet/in.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <assert.h>
#include "vector.h"
#include "units.h"


extern XFontStruct *vsmallFont, *vlargeFont;
GC brushedGC;

void LabelGraticule(Pixmap image, MapStruct * ms, GC gc)
{
    long GridU, ThinU;		/* Grid Spacing and meridian vertex spacing */
    long xu, xi, yu, start, end;
    int x, x1, y, len, xm = ms->WidthP, ym = ms->HeightP;
    float GridD;
    char label[10];


    if (ms->proj) return; // for now

    if (ms->GridSpace < DEF_GRID) {
	return;
    }
    if (ms->GridSpace == DEF_GRID)
	ms->GridSpace = DEF_GRID_SP;
    GridD = ms->GridSpace / PIX_PER_DEG(ms);
    if (GridD >= 12)
	GridD = 15;
    else if (GridD >= 8)
	GridD = 10;
    else if (GridD >= 3)
	GridD = 5;
    else if (GridD >= 1.5)
	GridD = 2;
    else if (GridD >= .75)
	GridD = 1;
    else if (GridD >= .35)
	GridD = .5;
    else if (GridD >= .15)
	GridD = .2;
    else
	GridD = .1;

    GridU = ThinU = DEG_TO_LU(GridD);

    /* Label Parallels. */
    start = ms->BoundsU.SLat - (ms->BoundsU.SLat % GridU);

    for (yu = start; yu < ms->BoundsU.NLat; yu += GridU) {
	y = MU_TO_PIX(ms, ms->CtrLatU - yu) + ms->HeightP / 2;
	if ((y < 0) || (y > ym - 10))
	    continue;
//      ImageLine(image, 4, y, 13, y, gc);
	if (GridD < 1)
	    sprintf(label, "%.1f", LU_TO_DEG(yu));
	else
	    sprintf(label, "%.0f", LU_TO_DEG(yu));
	ImageString(image, vsmallFont, 4, y + 10, label, gc);
    }

    /* Label Meridians */
    ym = 0;
    start = ms->BoundsU.WLon - (ms->BoundsU.WLon % GridU);
    if (ms->BoundsU.WLon > ms->CtrLonU)
	start -= DEG360_LU;
    end = ms->BoundsU.ELon;
    if (ms->BoundsU.ELon < ms->CtrLonU)
	end += DEG360_LU;
    yu = ms->BoundsU.NLat - PIX_TO_MU(ms, 12);


    Project(ms, yu, start, &x, &y);
    Project(ms, yu, start + GridU, &x1, &y);
    if ((x1 - x) < 20) return;

    for (xu = start; xu <= end; xu += GridU) {
	Project(ms, yu, xu, &x, &y);
	if ((x < 20) || (x > xm - 20))
	    continue;
//      ImageLine(image, x, 12, x, 15, gc);

xi = xu;
if (xi>180000000) xi -= DEG_TO_LU(360);
if (xi<-179000000) xi += DEG_TO_LU(360);
	if (GridD < 1)
	    sprintf(label, "%.1f", LU_TO_DEG(xi));
	else
	    sprintf(label, "%.0f", LU_TO_DEG(xi));

	x = x - (XTextWidth(vsmallFont, label, strlen(label)) / 2);
	ImageString(image, vsmallFont, x, 9, label, gc);
    }
}

int sort_by_pop(Place * a, Place * b)
{
    return (b->pop - a->pop);
}

void PlotLabel(Pixmap image, MapStruct * ms, long lat, long lon, Symbol * point,
	       Symbol * text, XFontStruct * font, char *label, Rectangle * labels, int num)
{
    int x, y, len, i, ch = 0, wid, ht;
    int labx[6], laby[6];
    int col;

    if (font == vsmallFont) {
	wid = 5;
	ht = 12;
    } else if (font == vlargeFont) {
	wid = 9;
	ht = 18;
    }
    len = strlen(label) * wid;

    Project(ms, lat, lon, &x, &y);
    if (point->brush != NULL) {
	ImageLine(image, x, y, x, y, point->brush);
	/* Locations of six preferences for labels */
	labx[0] = labx[1] = labx[2] = x + 6;
	labx[3] = labx[4] = labx[5] = x - len - 3;
	laby[0] = laby[3] = y - ht / 2;
	laby[1] = laby[4] = y - ht;
	laby[2] = laby[5] = y;
	if (x > ms->WidthP - len - 6)
	    ch = 3;
	/* Check for collisions with previously drawn labels */
	for (i = 0, col = 0; i < num; i++) {
	    if ((labx[ch] < labels[i].xr) && (labx[ch] + len > labels[i].xl) &&
	    (laby[ch] < labels[i].yb) && (laby[ch] + ht > labels[i].yt)) {
		/* collision has occured--pick next choice */
		ch++;
		col++;
		if (ch == 6) {
		    i = num;
		    ch = 0;
		} else
		    i = -1;
	    }
	}
	x = labx[ch];
	y = laby[ch];
    } else {
	/* No point to plot: position text centered over centroid */
	x -= len / 2;
	if (x < 0)
	    x = 0;
	if (x > ms->WidthP - len)
	    x = ms->WidthP - len;
	y -= ht / 2;
    }
    if (y < 0)
	y = 0;
    if (y > ms->HeightP - ht - 2)
	y = ms->HeightP - ht - 2;
    if (text != NULL && col<6) {
	ImageString(image, font, x, y+ht, label, text->brush);
	labels[num].xl = x;
	labels[num].xr = x + len;
	labels[num].yt = y;
	labels[num].yb = y + ht;
    }
    else {
	labels[num].xl = x;
	labels[num].xr = x;
	labels[num].yt = y;
	labels[num].yb = y;
    }
}

void LabelCities(Pixmap image, MapStruct * ms, Config * config)
{
    int places;
    int num, n, i = 0;
    void *pstart;
    char *p;
    Place pl, display[200];
    Rectangle labels[200];
    long psize;
    struct stat statbuf;
    char bufp[256];

    num = (ms->WidthP / 100) * (ms->HeightP / 100) + 10;
    n = num;

    sprintf(bufp,"%s/places.dat",get_vmap_root());
    places = open(bufp, O_RDONLY);
    if (fstat(places, &statbuf))
	return;
    psize = statbuf.st_size;

    pstart = mmap(0, psize, PROT_READ, MAP_SHARED, places, 0);

    (Place *)p = pstart + 64;		// skip header

    while (n) {
	i++;
	if (!(p < (char *)(pstart + psize)))
	    break;
	pl.lat = ntohl(MY_ALIGN(&((Place *) p)->lat));
	pl.lon = ntohl(MY_ALIGN(&((Place *) p)->lon));

	if (PtInRect(pl.lon, pl.lat, ms->BoundsU)) {
	    pl.pop = ntohl(MY_ALIGN(&((Place *) p)->pop));
	    memcpy(pl.st, ((Place *) p)->st, 2);

	    assert( (((Place *) p)->len - 15) < 1000);

	    memcpy(pl.name, ((Place *) p)->name, ((Place *) p)->len - 15);
	    pl.name[((Place *) p)->len - 15] = '\0';
	    display[num - n] = pl;
	    n--;
	}
	p += ((Place *) p)->len;
    }
    if (n)
	num -= n;
    munmap(pstart, psize);
    close(places);
    qsort((void *) display, num, sizeof(Place), sort_by_pop);

    for (n = 0; n < num; n++) {
	PlotLabel(image, ms, display[n].lat, display[n].lon, &config->point, &config->line,
	    n < 3 ? vlargeFont : vsmallFont, display[n].name, labels, n);
    }
}
